---------------------------------------------------------------------------
--standard Euphoria includes
include wildcard.e

--Bob's includes
include bobsgui.e
include bobsfile.e
include bobsmisc.e

--BAM player
include bamplay.e

---------------------------------------------------------------------------

constant false=0
constant true=1

---------------------------------------------------------------------------

constant MENU_QUIT=-1
constant MENU_ABORT=-2
constant MENU_OTHER=-3
constant MENU_TEST=-4

constant DEFAULT_PORT=#0388

---------------------------------------------------------------------------

integer current_port
integer made_changes made_changes=false

---------------------------------------------------------------------------

function update_menu(sequence menu,sequence data)

  --clean up menu 
  for i=1 to length(menu) do
    menu[i][2]=' '
    if data[i]=MENU_OTHER then
      menu[i]=sprintf("  Other         ????h  ",current_port)
    end if
  end for
 
  for i=1 to length(menu) do
    if current_port=data[i] then
      menu[i][2]='*'
      return(menu)
    end if
  end for

  for i=1 to length(menu) do
    if data[i]=MENU_OTHER then
      menu[i]=sprintf(" *Other         %04xh  ",current_port)
      return(menu)
    end if
  end for
   
  return(menu)
end function

---------------------------------------------------------------------------

procedure prompt_for_new_port()
  integer port
  port=prompt_for_hex("Base port number in hexidecmal",4,centered_block_by_size(20,6),{})
  if port then
    current_port=port
    made_changes=true
  end if
end procedure

---------------------------------------------------------------------------

function okay_to_quit()
  if made_changes then
    if not ask_yes_or_no(" Quit without saving changes? ",{BRIGHT_WHITE,RED}) then
      return(false)
    end if
  end if
  return(true)  
end function

---------------------------------------------------------------------------

procedure music_test()
  sequence block
  integer key
  integer vol

  vol=7

  set_fm_base_port(current_port)
  --set_mixer_base_port(current_port)

  if FM_SUPPORT then
    reset_fm()
    set_fm_melodic_mode()
    vol=get_fm_vol()
  end if

  --a few simple notes for testing the music
  set_bam_data(unpack_string({150,808530317,15856104,2013265978,825307533
  ,15856104,2013265978,842084749,15856104,2013265978,858861965,15856104
  ,2013265978,875639181,15856104,2013265978,892416397,15856104,2013265978
  ,909193613,15856104,2013265978,925970829,15856104,2013265978,942748045
  ,15856104,2013265978,2198873474,545329191,2183168016,780279937,270893600
  ,2165320833,545329189,2183168272,713105537,271024480,4261412864}))
  
  block=centered_block_by_size(44,6)
  push_text_block(block)
  bordered_wrap_box(block,0,sprintf("The sample song is now playing through port %04xh. Do you hear anything? You can adjust the volume with the arrow keys.",{current_port}))

  puts_at(block[1]+12,block[2]+4,sprintf(27&"[%s%s]"&26,{repeat('*',vol),repeat(' ',15-vol)}))

  while true do
    key=get_key()
    if key>=0 then
      if key=KEY_LEFT or key=KEY_DOWN then
        vol=bound(vol-1,0,15)
        set_fm_vol(vol)
        puts_at(block[1]+12,block[2]+4,sprintf(27&"[%s%s]"&26,{repeat('*',vol),repeat(' ',15-vol)}))
      elsif key=KEY_RIGHT or key=KEY_UP then
        vol=bound(vol+1,0,15)
        set_fm_vol(vol)
        puts_at(block[1]+12,block[2]+4,sprintf(27&"[%s%s]"&26,{repeat('*',vol),repeat(' ',15-vol)}))
      else
        exit
      end if
    end if
    advance_bam()
  end while

  reset_bam()
  reset_fm()
  pop_text_block()
end procedure

---------------------------------------------------------------------------

procedure write_soundset_ini()
  sequence ini
  integer found_match
  
  if file_exists("soundset.ini") then
    ini=explode(file_to_string("soundset.ini"),"\n")
  else
    ini={"# Base FM Port for applications that support BAM music"}  
  end if

  found_match=false  
  for i=1 to length(ini) do
    --check for match
    if compare("FMPORT=",upper(overflowing_slice(trim_whitespace(ini[i]),1,7,' ')))=0 then
      ini[i]=sprintf("FMPORT=%04x",{current_port})
      found_match=true
    end if
  end for
  
  if found_match=false then
    --no match was found, append.
    ini=append(ini,sprintf("FMPORT=%04x",{current_port}))
  end if

  string_to_file("soundset.ini",implode(ini,"\n"))
  
end procedure

---------------------------------------------------------------------------

procedure main_menu()
  sequence menu_text
  sequence menu_pointer
  sequence menu_block
  sequence menu_data

  integer key

  menu_text={"  Adlib         0388h  "
            ,"  Sound Blaster 0220h  "
            ,"  Sound Blaster 0240h  "
            ,"  Other         ????h  "
            ,"  Test Music           "
            ,"  Exit Without Saving  "
            ,"  Exit and Save"}
  menu_data={#0388
            ,#0220
            ,#0240
            ,MENU_OTHER
            ,MENU_TEST
            ,MENU_ABORT
            ,MENU_QUIT}

  menu_text=update_menu(menu_text,menu_data)

  menu_pointer=create_menu_pointer(1,1)
  menu_block=pick_block_coordinates(menu_text)

  push_text_block(menu_block)

  key=0
  while key != KEY_ESC do
    set_color({},false)
    draw_empty_box(menu_block)
    show_menu(constrict_block(menu_block,1),menu_pointer,menu_text,">",{})
    key=wait_key()
    if key=' ' then
      key=KEY_ENTER
    end if
    menu_pointer=evaluate_menu_keys(key,menu_pointer,menu_block[4],menu_text)
    if key=KEY_ENTER then
      if menu_data[menu_pointer[1]]>=0 then
        current_port=menu_data[menu_pointer[1]]
        made_changes=true
        menu_text=update_menu(menu_text,menu_data)
      end if
      if MENU_OTHER=menu_data[menu_pointer[1]] then
        prompt_for_new_port()
        menu_text=update_menu(menu_text,menu_data)
      end if
      if MENU_TEST=menu_data[menu_pointer[1]] then
        music_test()
      end if
      if MENU_ABORT=menu_data[menu_pointer[1]] then
        if okay_to_quit() then
          exit
        end if
      end if
      if MENU_QUIT=menu_data[menu_pointer[1]] then
        write_soundset_ini()
        exit
      end if
    end if
    if key=KEY_ESC then
      if okay_to_quit() then
        exit
      end if
    end if
  end while
  pop_text_block()

end procedure

---------------------------------------------------------------------------

function read_soundset_ini()

  sequence ini
  integer port

  if file_exists("soundset.ini")=false then
    --ini file does not exist, use default
    return(DEFAULT_PORT)
  end if

  --read file
  ini=explode(file_to_string("soundset.ini"),"\n")
  
  for i=1 to length(ini) do
    --check for match
    if compare("FMPORT=",upper(overflowing_slice(trim_whitespace(ini[i]),1,7,' ')))=0 then
      port=hex_from_string(overflowing_slice(ini[i],8,length(ini[i]),' '))
      return(port)
    end if
  end for
  
end function

---------------------------------------------------------------------------

enforce_program_dir()
init_text_gui()
current_port=read_soundset_ini()
main_menu()
cleanup_text_gui()

---------------------------------------------------------------------------

